AUTH_ERROR_CODE = 11001
SUBS_ERROR_CODE = 11011
UNSUBS_ERROR_CODE = 11012
RESUME_ERROR_CODE = 11031
RESUME_ERROR_CODE = 11032
MODE_ERROR_CODE = 12001
SUCCESS_CODE = 200
INVALID_CODE = -300
TOKEN_EXPIRED = -99
LIMIT_EXCEED_CODE = -99
LIMIT_EXCEED_MSG_5000 = "Please provide less than 5000 symbols"
TOKEN_EXPIRED_MSG = "Token is expired"
INVALID_TOKEN = "Please provide valid token"
SUCCESS = "ok"
ERROR = "error"
AUTH_SUCCESS = "Authentication done"
AUTH_FAIL = "Authentication failed"
SUBSCRIBE_SUCCESS = "Subscribed"
SUBSCRIBE_FAIL = "subscription failed"
UNSUBSCRIBE_SUCCESS = "Unsubscribed"
UNSUBSCRIBE_FAIL = "unsubscription failed"
LITE_MODE = "Lite Mode On"
FULL_MODE = "Full Mode On"
MODE_CHANGE_ERROR = "Mode change failed"
CHANNEL_PAUSED = "Channel Paused"
CHANNEL_RESUMED = "Channel Resumed"
CHANNEL_CHANGE_FAIL = "Mode change failed"
INVALID_SYMBOLS = "Please provide a valid symbol"
CONNECTION_CLOSED = "Connection Closed"
INDEX_DEPTH_ERROR_MESSAGE = "Index does not have market depth"
AUTH_TYPE = "cn"
SUBS_TYPE = "sub"
UNSUBS_TYPE = "unsub"
LITE_MODE_TYPE = "lit"
FULL_MODE_TYPE = "ful"
CH_PAUSE_TYPE = "cp"
CH_RESUME_TYPE = "cr"